@@ -66,6 +66,9 @@
 #define SD_FLAG_CMD_COW      0x02
 #define SD_FLAG_CMD_CACHE    0x04 /* Writeback mode for cache */
 #define SD_FLAG_CMD_DIRECT   0x08 /* Don't use cache */
+/* return something back while sending something to sheep */
+#define SD_FLAG_CMD_PIGGYBACK   0x10
+#define SD_FLAG_CMD_TGT   0x20
 
 #define SD_RES_SUCCESS       0x00 /* Success */
 #define SD_RES_UNKNOWN       0x01 /* Unknown error */
@@ -346,6 +349,7 @@ static const char *sd_strerror(int err)
 		{SD_RES_JOIN_FAILED, "Target node had failed to join sheepdog"},
 		{SD_RES_HALT, "Sheepdog is stopped serving IO request"},
 		{SD_RES_READONLY, "Object is read-only"},
+		{SD_RES_INODE_INVALIDATED, "Inode object is invalidated"},
 	};
 
 	for (i = 0; i < ARRAY_SIZE(errors); ++i) {
@@ -638,22 +642,37 @@ static int read_object(struct sheepdog_access_info *ai, char *buf, uint64_t oid,
 		       int copies, unsigned int datalen, uint64_t offset,
 		       int *need_reload);
 
-static int reload_inode(struct sheepdog_access_info *ai)
+static int reload_inode(struct sheepdog_access_info *ai, int is_snapshot)
 {
 	int ret, need_reload = 0;
 	char tag[SD_MAX_VDI_TAG_LEN];
 	uint32_t vid;
 
-	memset(tag, 0, sizeof(tag));
+	if (is_snapshot) {
+		memset(tag, 0, sizeof(tag));
 
-	ret = find_vdi_name(ai, ai->inode.name, CURRENT_VDI_ID, tag, &vid, 0);
-	if (ret)
-		return -1;
+		ret = find_vdi_name(ai, ai->inode.name, CURRENT_VDI_ID, tag,
+				    &vid, 0);
+		if (ret)
+			return -1;
 
-	ret = read_object(ai, (char *)&ai->inode, vid_to_vdi_oid(vid),
-			  ai->inode.nr_copies, SD_INODE_SIZE, 0, &need_reload);
-	if (ret)
-		return -1;
+		ret = read_object(ai, (char *)&ai->inode, vid_to_vdi_oid(vid),
+				  ai->inode.nr_copies,
+				  offsetof(struct sheepdog_inode, data_vdi_id),
+				  0, &need_reload);
+		if (ret)
+			return -1;
+	} else {
+		ret = read_object(ai, (char *)&ai->inode,
+				  vid_to_vdi_oid(ai->inode.vdi_id),
+				  ai->inode.nr_copies, SD_INODE_SIZE, 0,
+				  &need_reload);
+		if (ret)
+			return -1;
+	}
+
+	ai->min_dirty_data_idx = UINT32_MAX;
+	ai->max_dirty_data_idx = 0;
 
 	return 0;
 }
@@ -692,6 +711,7 @@ static int read_write_object(struct sheepdog_access_info *ai, char *buf,
 	hdr.data_length = datalen;
 	hdr.offset = offset;
 	hdr.copies = copies;
+	hdr.flags |= SD_FLAG_CMD_TGT;
 
 	ret = do_req(ai, (struct sheepdog_req *)&hdr, buf, &wlen, &rlen);
 	if (ret) {
@@ -704,6 +724,8 @@ static int read_write_object(struct sheepdog_access_info *ai, char *buf,
 		return 0;
 	case SD_RES_INODE_INVALIDATED:
 		dprintf("inode object is invalidated\n");
+		*need_reload = 2;
+		return 0;
 	case SD_RES_READONLY:
 		*need_reload = 1;
 		return 0;
@@ -766,7 +788,7 @@ static int sd_sync(struct sheepdog_access_info *ai)
 
 static int update_inode(struct sheepdog_access_info *ai)
 {
-	int ret = 0;
+	int ret = 0, need_reload_inode = 0;
 	uint64_t oid = vid_to_vdi_oid(ai->inode.vdi_id);
 	uint32_t min, max, offset, data_len;
 
@@ -776,16 +798,29 @@ static int update_inode(struct sheepdog_access_info *ai)
 	if (max < min)
 		goto end;
 
+	goto update;
+
+reload:
+	reload_inode(ai, 0);
+	need_reload_inode = 0;
+
+update:
 	offset = sizeof(ai->inode) - sizeof(ai->inode.data_vdi_id) +
 		min * sizeof(ai->inode.data_vdi_id[0]);
 	data_len = (max - min + 1) * sizeof(ai->inode.data_vdi_id[0]);
 
 	ret = write_object(ai, (char *)&ai->inode + offset, oid,
 			   ai->inode.nr_copies, data_len, offset,
-			   0, 0, 0, NULL);
+			   0, 0, 0, &need_reload_inode);
 	if (ret < 0)
 		eprintf("sync inode failed\n");
 
+	if (need_reload_inode) {
+		dprintf("reloading inode is required in the path"
+			" of update_inode()\n");
+		goto reload;
+	}
+
 end:
 	ai->min_dirty_data_idx = UINT32_MAX;
 	ai->max_dirty_data_idx = 0;
@@ -799,18 +834,16 @@ static int is_refresh_required(struct sheepdog_access_info *ai)
  * 1: refresh is required
  */
 {
-	uint64_t first_oid = vid_to_data_oid(ai->inode.vdi_id, 0);
+	uint64_t inode_oid = vid_to_vdi_oid(ai->inode.vdi_id);
 	char dummy;
 	int need_reload_inode = 0;
 
 	/*
 	 * Check inode of this tgtd is invaldiated or not.
-	 * Reading which object isn't a problem. If the inode object is
-	 * invalidated, sheep returns SD_RES_INODE_INVALIDATED even if the
-	 * object doesn't exist. So we read the first one.
+	 * The inode object is the only one object which always exists.
 	 */
 
-	read_object(ai, &dummy, first_oid, ai->inode.nr_copies, sizeof(dummy),
+	read_object(ai, &dummy, inode_oid, ai->inode.nr_copies, sizeof(dummy),
 		    0, &need_reload_inode);
 
 	return need_reload_inode;
@@ -837,7 +870,7 @@ reload_in_read_path:
 	pthread_rwlock_unlock(&ai->inode_lock); /* unlock current read lock */
 
 	pthread_rwlock_wrlock(&ai->inode_lock);
-	ret = reload_inode(ai);
+	ret = reload_inode(ai, 0);
 	if (ret) {
 		eprintf("failed to reload in read path\n");
 		goto out;
@@ -895,7 +928,13 @@ retry:
 					   old_oid, flags, &need_reload_inode);
 			if (!ret) {
 				if (need_reload_inode) {
-					ret = reload_inode(ai);
+					/* If need_reload_inode is 1,
+					 * snapshot was created.
+					 * If it is 2, inode object is
+					 * invalidated
+					 */
+					ret = reload_inode(ai,
+						   need_reload_inode == 1);
 					if (!ret)
 						goto retry;
 				}
@@ -1157,6 +1196,7 @@ trans_to_expect_nothing:
 	ai->min_dirty_data_idx = UINT32_MAX;
 	ai->max_dirty_data_idx = 0;
 
+retry:
 	ret = read_object(ai, (char *)&ai->inode, vid_to_vdi_oid(vid),
 			  0, SD_INODE_SIZE, 0, &need_reload);
 	if (ret)
