@@ -72,6 +72,19 @@ extern void log_error(const char *fmt, ...)
 extern void log_debug(const char *fmt, ...)
 	__attribute__ ((format (printf, 1, 2)));
 
+#ifdef NO_LOGGING
+#define eprintf(fmt, args...)						\
+do {									\
+	fprintf(stderr, "%s: " fmt, program_name, ##args);		\
+} while (0)
+
+#define dprintf(fmt, args...)						\
+do {									\
+	if (debug)							\
+		fprintf(stderr, "%s %d: " fmt,				\
+			__FUNCTION__, __LINE__, ##args);		\
+} while (0)
+#else
 #define eprintf(fmt, args...)						\
 do {									\
 	log_error("%s(%d) " fmt, __FUNCTION__, __LINE__, ##args);	\
@@ -82,5 +95,6 @@ do {									\
 	if (unlikely(is_debug))						\
 		log_debug("%s(%d) " fmt, __FUNCTION__, __LINE__, ##args); \
 } while (0)
+#endif
 
 #endif	/* LOG_H */
